/* fakefibers.h - Emulated fibers and coroutines using GCC extensions
 *
 * Copyright (C) 2009 Timo Teräs <timo.teras@iki.fi>
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 or later as
 * published by the Free Software Foundation.
 *
 * See http://www.gnu.org/ for details.
 */

/* For the basic concepts, go and read:
 * http://www.chiark.greenend.org.uk/~sgtatham/coroutines.html
 * http://www.sics.se/~adam/pt/
 *
 * This implements same idea with GNU C extensions:
 * http://gcc.gnu.org/onlinedocs/gcc/Labels-as-Values.html
 * http://gcc.gnu.org/onlinedocs/gcc/Local-Labels.html
 *
 * continuation_t and co_* implement a simple statemachine style coroutine,
 * which can yield and be resumed with a context.

 * struct async_ctx and async_* are specialized coroutines with
 * defined constructor and virtualized process function. The process function
 * returns -EAGAIN while it has blocked (and scheduled some event
 * that resumes processing) or some other value when it exists and
 * has cleaned up. Support for cancelling such a cofunction is optional.
 * Nesting is possible, but each nesting level requires an additional
 * function call on resume, so deep nesting is not recommended.
 */

#ifndef FAKEFIBERS_H
#define FAKEFIBERS_H

#include <errno.h>
#include <stddef.h>
#include "ev.h"

#ifndef offsetof
#ifdef __compiler_offsetof
#define offsetof(TYPE,MEMBER) __compiler_offsetof(TYPE,MEMBER)
#else
#define offsetof(TYPE, MEMBER) ((size_t) &((TYPE *)0)->MEMBER)
#endif
#endif

#ifndef container_of
#define container_of(ptr, type, member) ({                      \
        const typeof( ((type *)0)->member ) *__mptr = (ptr);    \
        (type *)( (char *)__mptr - offsetof(type,member) );})
#endif

struct co_context {
	ptrdiff_t pos;
};

#define CONTINUATION_INITIALIZER (struct co_context) { 0 }

#define coroutine_attribute	__attribute__((__noinline__))

#define co_start(ctx) \
	struct co_context *__cont = ctx;				\
	goto *(&&__cont_start + __cont->pos);				\
	__cont_start:

#define co_yield(x)							\
	({								\
		__label__ __cont_resume;				\
		__cont->pos = &&__cont_resume - &&__cont_start;		\
		return (x);						\
	__cont_resume: ;						\
	})

#define co_end(x)							\
	({								\
		__label__ __cont_end;					\
		__cont->pos = &&__cont_end - &&__cont_start;		\
	__cont_end:							\
		return (x);						\
	})

#define async_attribute		coroutine_attribute

struct async_thread {
	struct ev_loop *loop;
	struct async_context *resume_context;
	int cancel_signal;
};

struct async_context {
	struct async_thread *thread;
	struct async_context *parent;
	int (*resume)(struct async_context *);
	struct co_context co;
	int retval;
};

#define async_start(ctx) \
	struct async_context *__async_self = ctx;			\
	co_start(&__async_self->co)

#define async_call(async_func, child_ctx)				\
	({								\
		struct async_context *__child = (child_ctx);		\
		int rc;							\
		__child->parent = NULL;					\
		__child->resume = async_func;				\
		__child->co = CONTINUATION_INITIALIZER;			\
		rc = async_func(__child);				\
		if (rc == -EAGAIN) {					\
			__child->parent = __async_self;			\
			co_yield(-EAGAIN);				\
			rc = (child_ctx)->retval;			\
		}							\
		rc;							\
	})

#define async_launch(async_func, child_ctx)				\
	({								\
		struct async_context *__child = (child_ctx);		\
		__child->parent = NULL;					\
		__child->resume = async_func;				\
		__child->co = CONTINUATION_INITIALIZER;			\
		async_func(__child);					\
	})

#define async_end(rc)							\
	__async_self->retval = rc;					\
	if (__async_self->parent != NULL)				\
		return __async_self->parent->resume(__async_self->parent); \
	return __async_self->retval;

void async_resume_from_libev(EV_P_ ev_watcher *w, int revents);

#define async_sleep(timer, timeout_sec)					\
	({								\
		ev_timer *__timer = timer;				\
		ev_timer_init(__timer, ((void (*)(EV_P_ ev_timer *, int)) async_resume_from_libev), timeout_sec, 0.); \
		__timer->data = __async_self;				\
		ev_timer_start(EV_DEFAULT, __timer);			\
		co_yield(-EAGAIN);					\
		ev_timer_stop(EV_DEFAULT, timer);			\
	})

#endif
